--------------------------------------------------
-- ArekosGratz, Author: Areko                   --
--------------------------------------------------
local ADDONNAME, ArekosGratz = ...

LibStub("AceAddon-3.0"):NewAddon(ArekosGratz, ADDONNAME, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale("ArekosGratz")

local LSM = LibStub("LibSharedMedia-3.0")

-- GLOBALS: LibStub
local _G = _G

local string, strjoin, tostring = string, strjoin, tostring
local type, select, gsub, random, ceil, bit = type, select, gsub, random, math.ceil, bit
local tinsert, tremove, pairs, ipairs, unpack, next, setmetatable, wipe, rawget, rawset = tinsert, tremove, pairs, ipairs, unpack, next, setmetatable, wipe, rawget, rawset

local lower = string.utf8lower or string.lower

local NEAR_TYPES = {
    NONE    = 0,    -- L["NONE"]    = "None"
    SAY     = 1,    -- L["SAY"]     = "Say"
    WHISPER = 2,    -- L["WHISPER"] = "Whisper"
}
local NEAR_TYPES_VALUES = {}
for name, value in pairs(NEAR_TYPES) do
    NEAR_TYPES_VALUES[value] = L[name]
end
local MESSAGE_FLAGS = {
    NONE        = 0,
    NEAR        = 1,
    GROUP       = 2,
    GUILD       = 4,
    INSTANCE    = 8,

    GRATZ       = 16,
    THANKS      = 32,
}

local messageDefaults = {
    exists = false,
    isIntern = false,
    removeRealmName = false,
    capitalizeFirstChar = false,
    onlyFor_HasBattleNetTag = false,
    notFor_HasBattleNetTag = false,
    onlyFor = {
        -- [#] = "name-realm" -> Playername
        -- [#] = "Name#1234" -> Battle.net Tag
    },
    notFor = {
        -- same as onlyFor
    },
}
local defaults = {
    global = {
        gzTriggersImported = false,
        gzTriggers = {
            gz = true,
            gw = true,
            gg = true,
            gzi = true,
            grats = true,
            gratz = true,
            salute = true,
            congrat = true,
            goncrats = true,
            dingrats = true,
            gradulate = true,
            condolence = true,
            coragulate = true,
            congrations = true,
            gratulation = true,
            congrazzles = true,
            congratsory = true,
            ["glückwunsch"] = true,
            felicitation = true,
            concragulate = true,
            congradulate = true,
            comendations = true,
            congraduation = true,
            confagulations = true,
            congratulation = true,
            congratulating = true,
            congratulations = true,
            concrapulations = true,
        },
        scheduleTimer = {
            ['*'] = 0.5,
            sendMessageDelay = 1,
            cleanMessageDelay = 1,
        },
    },
    profile = {
        protectAFK = true,
        intellectual_random = false,
        congratulations = {
            guild = false,
            party = false,
            ignoreInstanceGroup = false,
            near = NEAR_TYPES.NONE,
            guildInOtherChannel = false,
            minWaitTime = 3,
            maxWaitTime = 5,
        },
        thanks = {
            guild = false,
            party = false,
            instance = false,
            near = NEAR_TYPES.NONE,
            minWaitTime = 15,
            maxWaitTime = 20,
            maxExtraThanksTrys = 1,
        },
        chance = {
            ['**'] = {
                guild = 1,
                party = 1,
                near = 1,
            },
            -- congratulations = {},
            thanks = {
                instance = 1,
            },
        },
        playSound = false,
        soundChannel = "Master",
        sound = "",
        gratzKeywords = nil,
        thxKeywords = nil,
        keywordsImported = false,
        messageTHX = {
            ['**'] = messageDefaults,
        },
        messageGZ = {
            ['**'] = messageDefaults,
        },
    }
}

local get, del
do -- TablePool
    local tablePool = setmetatable({}, { __mode="k" })
    get = function()
        local t = next(tablePool)
        if t then
            tablePool[t] = nil
            return t
        else
            return {}
        end
    end
    del = function(t)
        setmetatable(t, nil)    -- removes metatable
        wipe(t)
        tablePool[t] = true
    end
end

local HOMEREALM = _G.GetRealmName():gsub("%s+", "")
ArekosGratz.thanksPendingList = {}
ArekosGratz.gratzPendingList = {}

function ArekosGratz:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("ArekosGratzDB", defaults, true)

    self.db.RegisterCallback(self, "OnProfileChanged")
    self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
    self.db.RegisterCallback(self, "OnDatabaseShutdown")

    if not self.db.global.gzTriggersImported then
        local gratzList = {
            global = {
                triggers = nil,
            },
        }
        local gratzDB = LibStub("AceDB-3.0"):New("GratzDB", gratzList, true)
        if type(gratzDB.global.triggers) == "table" then
            for _, trigger in pairs(gratzDB.global.triggers) do
                self.db.global.gzTriggers[ lower(trigger) ] = true
            end
        end
        self.db.global.gzTriggersImported = true
    end

    -- We call this for import of old settings on profile level and message* db integrity check
    self:OnProfileChanged("OnProfileChanged", self.db, self.db:GetCurrentProfile() )

    self:RegisterChatCommand("arekosgratz", "ChatCommand")
    self:RegisterChatCommand("ag", "ChatCommand")

    self:RegisterConfig()

    self.playername = _G.UnitName("player").."-".._G.GetRealmName():gsub("%s", "")
    self.playernameLower = lower(self.playername)
    self:GuildUpdate()
end

function ArekosGratz:OnProfileChanged(event, db, profile)
    if not db.profile.keywordsImported then
        --[===[@debug@
        self:Print("Start import of old settings: ", profile or db:GetCurrentProfile() )
        --@end-debug@]===]
        if type(db.profile.gratzKeywords) == "table" then
            for key, data in pairs(db.profile.gratzKeywords) do
                key = key:trim()
                db.profile.messageGZ[key].exists = true
                for id, value in ipairs(data.onlyFor) do
                    db.profile.messageGZ[key].onlyFor[id] = lower( self:UniformPlayerName( tostring(value) ) )
                end
                for id, value in ipairs(data.notFor) do
                    db.profile.messageGZ[key].notFor[id] = lower( self:UniformPlayerName( tostring(value) ) )
                end
            end
        end
        if type(db.profile.thxKeywords) == "table" then
            for key, data in pairs(db.profile.thxKeywords) do
                key = key:trim()
                db.profile.messageTHX[key].exists = true
                for id, value in ipairs(data.onlyFor) do
                    db.profile.messageTHX[key].onlyFor[id] = lower( self:UniformPlayerName( tostring(value) ) )
                end
                for id, value in ipairs(data.notFor) do
                    db.profile.messageTHX[key].notFor[id] = lower( self:UniformPlayerName( tostring(value) ) )
                end
            end
        end
        db.profile.keywordsImported = true
    end

    -- aka message* db integrity check
    self:CheckForBattleNetTags(db)

    self:ResetOptionsCache()

    if self:IsEnabled() then
        -- Profile changed and we are already enabled, so check if battle.net tag indexing is needed

        if self:HasBattleNetTags() then
            self:InitializeBattleNetTagIndex()
        else
            self:DisableBattleNetTagIndex()
        end
    end
end

function ArekosGratz:OnDatabaseShutdown(event, db)
    -- clean up unused triggers
    for key, inUse in pairs(db.global.gzTriggers) do
        if not inUse and not defaults.global.gzTriggers[key] then
            db.global.gzTriggers[key] = nil
        end
    end
end

function ArekosGratz:OnEnable()
    self:GuildUpdate()
    self:RegisterEvent("PLAYER_GUILD_UPDATE", "GuildUpdate")

    if self:HasBattleNetTags() then
        self:InitializeBattleNetTagIndex()
    end

    self:RegisterEvent("ACHIEVEMENT_EARNED")
    self:RegisterEvent("CHAT_MSG_ACHIEVEMENT", "ACHIEVEMENT_EARNED_BY_OHTER_PLAYER")
    self:RegisterEvent("CHAT_MSG_GUILD_ACHIEVEMENT", "ACHIEVEMENT_EARNED_BY_OHTER_PLAYER")
end

function ArekosGratz:OnDisable()
end

function ArekosGratz:GuildUpdate(event)
    local guild, _, _, realm = _G.GetGuildInfo("player")
    self.guildname = ( type(guild) == "string" and guild:len() > 0 ) and self:UniformGuildName(guild, realm or nil) or ""
    self.guildnameLower = lower(self.guildname)
end

do-- Battle.net tag index
    local lastDB_HasBattleNetTags = false
    local index = setmetatable({
        --[BNTag] = "name-realm"
    }, {
        __index = function (t, k)
            return k
        end
    } )

    local function checkFriendList()
        local numFriends, numOnline = _G.BNGetNumFriends()

        for friendIndex = 1, numOnline do
            local battleTag = select(3, _G.BNGetFriendInfo(friendIndex) )

            -- NOTE: When we encounter user where have friends which use multiple wow accounts in same time:
            -- Find a good way to handle more than one wow account on a BattleNet profile
            for toonIndex = 1, _G.BNGetNumFriendToons(friendIndex) do
                local toonName, client, realmName = select(2, _G.BNGetFriendToonInfo(friendIndex, toonIndex) )

                if _G.BNET_CLIENT_WOW == client then
                    index[lower(battleTag)] = lower( ArekosGratz:UniformPlayerName(toonName, realmName) )
                end
            end
        end
    end

    function ArekosGratz:InitializeBattleNetTagIndex()
        --[===[@debug@
        self:Print("InitializeBattleNetTagIndex")
        --@end-debug@]===]

        self:RegisterEvent("BN_FRIEND_TOON_ONLINE")
        self:RegisterEvent("BN_ACCOUNT_TOON_ONLINE", "BN_FRIEND_TOON_ONLINE")

        checkFriendList()
        --[===[@debug@
        for bn, name in pairs(index) do
            self:Printf("Toons: %s=%s", bn, name)
        end
        --@end-debug@]===]
    end

    function ArekosGratz:DisableBattleNetTagIndex()
        --[===[@debug@
        self:Print("DisableBattleNetTagIndex")
        --@end-debug@]===]
        self:UnregisterEvent("BN_FRIEND_TOON_ONLINE")
        self:UnregisterEvent("BN_ACCOUNT_TOON_ONLINE")
    end

    -- BN_FRIEND_TOON_OFFLINE: toonID
    -- BN_FRIEND_TOON_ONLINE: toonID

    function ArekosGratz:BN_FRIEND_TOON_ONLINE(event, toonID)
        if not toonID then return end

        local toonName, client, realmName = select(2, _G.BNGetToonInfo(toonID) )

        --[===[@debug@
        self:Printf("BN_FRIEND_TOON_ONLINE: %s-%s is client %s, id:%s", toonName or "<nil>", tostring(realmName), client or "<nil>", tostring(toonID))
        --@end-debug@]===]

        if _G.BNET_CLIENT_WOW == client then
            -- we know now that a new toon is online so lets check all your online friends,
            -- because we don't get the BattleTag from a toon..
            checkFriendList()
        end
    end

    local scopes = { "messageGZ", "messageTHX" }
    local filters = { "notFor", "onlyFor" }

    function ArekosGratz:CheckForBattleNetTags(db)
        local db = db or self.db
        lastDB_HasBattleNetTags = false

        for _, scope in pairs(scopes) do
            for message, data in pairs(db.profile[scope]) do
                if data.exists then
                    for _, filter in pairs(filters) do
                        local hasBattleNetTag = false
                        for _, value in pairs(data[filter]) do
                            if value:find('#') then
                                hasBattleNetTag = true
                                lastDB_HasBattleNetTags = true
                                break
                            end
                        end
                        db.profile[scope][message][filter.."_HasBattleNetTag"] = hasBattleNetTag
                    end
                else
                    db.profile[scope][message] = nil
                end
            end
        end
        --[===[@debug@
        self:Print("CheckForBattleNetTags:", tostring(lastDB_HasBattleNetTags) )
        --@end-debug@]===]
    end

    function ArekosGratz:HasBattleNetTags()
        return lastDB_HasBattleNetTags
    end

    function ArekosGratz:GetToonNameFromBattleNetTag(BNTag)
        return index[ lower(BNTag) ]
    end

    --@alpha@
    function ArekosGratz:GetBattleNetTagIndex()
        return index
    end
    --@end-alpha@
end

function ArekosGratz:ChatCommand(input)
    local a1, lastPos = self:GetArgs(input, 1)

    if not a1 or a1:len() < 1 or a1 == "op" or a1 == "options" then
        _G.InterfaceOptionsFrame_OpenToCategory(L["Areko's Gratz"])
        _G.InterfaceOptionsFrame_OpenToCategory(L["Areko's Gratz"]) -- Note: remove this when not any more needed
    else
        if a1 == "config" then
            LibStub("AceConfigCmd-3.0").HandleCommand(self, "ag config", ADDONNAME.."OptionsMain", input:sub(lastPos) )
        elseif a1 == "bn" or a1 == "bnet" or a1 == "battlenet" then
            local a2, lastPos = self:GetArgs(input, 1, lastPos)
            local count
            if a2 == "all" then
                count = _G.BNGetNumFriends()
            else
                self:Printf(L["/ag %s all      - Will list offline friends too"], a1)
                count = select(2, _G.BNGetNumFriends() )
            end

            for friendIndex = 1, count do
                local presenceID, presenceName, battleTag, isBattleTagPresence = _G.BNGetFriendInfo(friendIndex)

                if type(battleTag) == "string" then
                    self:Printf(L["%s%s|r BattleTag is %s%s|r"], _G.FRIENDS_PRESENCE_COLOR_CODE or "|cff7c848a", presenceName, _G.FRIENDS_BNET_NAME_COLOR_CODE or "|cff82c5ff", battleTag)
                end
            end
        elseif a1 == "rnd" or a1 == "random" then
            self.db.profile.intellectual_random = not self.db.profile.intellectual_random
            self:Print(L["'Predictable random' messages %s"]:format( self.db.profile.intellectual_random and L["are enabled"] or L["are disabled"] ))
        else
            self:Print(L["Usage: /ag <command> or /arekosgratz <command>"])
            self:Print(L["/ag op|options    - Open Options"])
            self:Print(L["/ag bn|bnet       - List BattleTags of your friends"])
            self:Print(L["/ag rnd|random    - Toggle the 'predictable random' setting"])
            self:Print(L["/ag config        - Change settings"])
        end
    end
end

do-- HasChance
    local chanceCounts = setmetatable({}, {
        __index = function(t, k)
            local mask = setmetatable({}, {
                __index = function(t,k)
                    local data = {
                        hit = 1,
                        miss = 0,
                    }
                    rawset(t, k, data)
                    return data
                end
            })
            rawset(t, k, mask)
            return mask
        end,
    })

    function ArekosGratz:HasChance(chance, mask)
        chance = chance * 100

        if chance >= 100 then
            return true
        end

        if mask then
            local data = chanceCounts[mask][chance]
            local result
            local currentHitChance = 1 / ( data.hit + data.miss ) * data.hit

            if currentHitChance < chance then
                result = true
            else
                result = random(0, 100) <= chance
            end

            if result then
                data.hit = data.hit + 1
            else
                data.miss = data.miss + 1
            end

            --[===[@debug@
            self:Printf("Chance[%s]: %s < %s forced(), result: %s", tostring(mask), currentHitChance, chance, tostring(currentHitChance < chance), tostring(result) )
            --@end-debug@]===]

            return result
        else
            return random(0, 100) <= chance
        end
    end
end

--[[
ACHIEVEMENT_EARNED, a1=achievementid
CHAT_MSG_ACHIEVEMENT, a1=message,a2=playername
CHAT_MSG_GUILD_ACHIEVEMENT, a1=message,a2=playername,a12=GUID


CHAT_MSG_GUILD, a1=message,a2=user
CHAT_MSG_PARTY, a1=message,a2=user
CHAT_MSG_PARTY_LEADER, a1=message,a2=user
CHAT_MSG_RAID, a1=message,a2=user
CHAT_MSG_RAID_LEADER, a1=message,a2=user
CHAT_MSG_SAY, a1=message,a2=user
CHAT_MSG_WHISPER, a1=message,a2=user
]]

function ArekosGratz:IsInMyGuild(name)
    -- NOTE: maybe add a better way to test this
    return _G.UnitIsInMyGuild(name)
end

function ArekosGratz:IsInMyGroup(name)
    -- NOTE: check for better way
    return _G.UnitInParty(name) or _G.UnitInRaid(name)
end

function ArekosGratz:IsMe(name)
    local playername = lower( self:UniformPlayerName(name) )

    return playername == self.playernameLower
end

function ArekosGratz:IsMyGuild(guild)
    local guildname = lower( self:UniformGuildName(guild) )

    return guildname == self.guildnameLower
end

function ArekosGratz:IAmAFK()
    return _G.UnitIsAFK("player")
end

function ArekosGratz:GetGroupChannel(ignoreInstanceGroup)
    local result = ""
    if ignoreInstanceGroup then
        result = _G.IsInRaid(_G.LE_PARTY_CATEGORY_HOME) and "RAID" or _G.IsInGroup(_G.LE_PARTY_CATEGORY_HOME) and "PARTY" or ""
    else
        result = _G.IsInGroup(_G.LE_PARTY_CATEGORY_INSTANCE) and "INSTANCE_CHAT" or ( _G.IsInRaid() and "RAID" or "PARTY" )
    end

    return result
end

local nearOrGroupFlag = MESSAGE_FLAGS.GROUP + MESSAGE_FLAGS.NEAR
function ArekosGratz:ACHIEVEMENT_EARNED_BY_OHTER_PLAYER(event, message, playername)
    if self:IAmAFK() and self.db.profile.protectAFK then return end
    if type(playername) ~= "string" or playername:len() < 1 then return end
    local guildname = playername -- When guild achievement, `playername` is the name of the guild
    playername = lower( self:UniformPlayerName( playername ) )
    if self:IsMe(playername) or self:IsMyGuild(guildname) then return end

    if event == "CHAT_MSG_ACHIEVEMENT" then
        if not self.db.profile.congratulations.guildInOtherChannel and self:IsInMyGuild(playername) then
            return
        end

        if self:IsInMyGroup(playername) then
            if self.db.profile.congratulations.party then
                self.gratzPendingList[playername] = bit.bor(MESSAGE_FLAGS.GROUP, (self.gratzPendingList[playername] or MESSAGE_FLAGS.NONE) )

                if self.TimeLeft(self.gratzTimerNearGroup) <= 0 then
                    self.gratzTimerNearGroup = self:ScheduleTimer("GratzForAchievement", random(self.db.profile.congratulations.minWaitTime, self.db.profile.congratulations.maxWaitTime), nearOrGroupFlag )
                end
            end
        else
            if self.db.profile.congratulations.near ~= NEAR_TYPES.NONE then
                self.gratzPendingList[playername] = bit.bor(MESSAGE_FLAGS.NEAR, (self.gratzPendingList[playername] or MESSAGE_FLAGS.NONE) )

                if self.TimeLeft(self.gratzTimerNearGroup) <= 0 then
                    self.gratzTimerNearGroup = self:ScheduleTimer("GratzForAchievement", random(self.db.profile.congratulations.minWaitTime, self.db.profile.congratulations.maxWaitTime), nearOrGroupFlag )
                end
            end
        end
    elseif event == "CHAT_MSG_GUILD_ACHIEVEMENT" then
        if self.db.profile.congratulations.guild then
            self.gratzPendingList[playername] = bit.bor(MESSAGE_FLAGS.GUILD, (self.gratzPendingList[playername] or MESSAGE_FLAGS.NONE) )

            if self.TimeLeft(self.gratzTimerGuild) <= 0 then
                self.gratzTimerGuild = self:ScheduleTimer("GratzForAchievement", random(self.db.profile.congratulations.minWaitTime, self.db.profile.congratulations.maxWaitTime), MESSAGE_FLAGS.GUILD )
            end
        end
    end
end

function ArekosGratz:GratzForAchievement(flag)
    --[===[@debug@
    local start = _G.GetTime()
    --@end-debug@]===]
    if bit.band(flag, MESSAGE_FLAGS.GUILD) ~= 0 and self:HasChance(self.db.profile.chance.congratulations.guild, MESSAGE_FLAGS.GUILD+MESSAGE_FLAGS.GRATZ) then
        local guild = get()
        local found = 0

        for name, flags in pairs(self.gratzPendingList) do
            if bit.band(flags, MESSAGE_FLAGS.GUILD) ~= 0 then
                guild[name] = true
                found = found + 1
                self.gratzPendingList[name] = bit.bxor(flags, MESSAGE_FLAGS.GUILD)
            end
        end

        if found > 0 then
            local messages = self:GetMessages(guild, "messageGZ", found)
            if type(messages) == "table" then
                local i = 0
                for _, message in pairs(messages) do
                    if i == 0 then
                        self:SendChatMessage(message, "GUILD")
                    else
                        self:ScheduleTimer("SendChatMessage", self.db.global.scheduleTimer.sendMessageDelay*i, message, "GUILD" )
                    end
                    i = i + 1
                end
                del(messages)
            elseif type(messages) == "string" then
                self:SendChatMessage(messages, "GUILD")
            end
        end
        del(guild)
    end
    if bit.band(flag, nearOrGroupFlag) ~= 0 then
        local group = get()
        local groupFound = 0
        local near = get()
        local nearFound = 0

        local groupAllowed
        local nearAllowed

        for name, flags in pairs(self.gratzPendingList) do
            if bit.band(flags, MESSAGE_FLAGS.GROUP) ~= 0 then
                if groupAllowed == nil then
                    groupAllowed = self:HasChance(self.db.profile.chance.congratulations.party, MESSAGE_FLAGS.GROUP+MESSAGE_FLAGS.GRATZ)
                end
                if groupAllowed then
                    group[name] = true
                    groupFound = groupFound + 1
                    self.gratzPendingList[name] = bit.bxor(flags, MESSAGE_FLAGS.GROUP)
                end
            end
            if bit.band(flags, MESSAGE_FLAGS.NEAR) ~= 0 then
                if nearAllowed == nil then
                    nearAllowed = self:HasChance(self.db.profile.chance.congratulations.near, MESSAGE_FLAGS.NEAR+MESSAGE_FLAGS.GRATZ)
                end
                if nearAllowed then
                    near[name] = true
                    nearFound = nearFound + 1
                    self.gratzPendingList[name] = bit.bxor(flags, MESSAGE_FLAGS.NEAR)
                end
            end
        end

        if groupFound > 0 then
            local channel = self:GetGroupChannel(self.db.profile.congratulations.ignoreInstanceGroup)

            if type(channel) == "string" and channel:len() > 1 then
                local messages = self:GetMessages(group, "messageGZ", groupFound)

                if type(messages) == "table" then
                    local i = 0
                    for _, message in pairs(messages) do
                        if i == 0 then
                            self:SendChatMessage(message, channel)
                        else
                            self:ScheduleTimer("SendChatMessage", self.db.global.scheduleTimer.sendMessageDelay*i, message, channel)
                        end
                        i = i + 1
                    end
                    del(messages)
                elseif type(messages) == "string" then
                    self:SendChatMessage(messages, channel)
                end
            end
        end
        if nearFound > 0 then
            if self.db.profile.thanks.near == NEAR_TYPES.WHISPER then
                for name in pairs(near) do
                    local nameTable = get()
                    nameTable[name] = true
                    local messages = self:GetMessages(nameTable, "messageGZ", nearFound)
                    if type(messages) == "table" then
                        local message = messages[random(1, #messages)]

                        self:SendChatMessage(message, "WHISPER", name)
                        del(messages)
                    elseif type(messages) == "string" then
                        self:SendChatMessage(messages, "WHISPER", name)
                    end
                    del(nameTable)
                end
            elseif self.db.profile.thanks.near == NEAR_TYPES.SAY then
                local messages = self:GetMessages(near, "messageGZ", nearFound)
                if type(messages) == "table" then
                    local i = 0
                    for _, message in pairs(messages) do
                        if i == 0 then
                            self:SendChatMessage(message, "SAY")
                        else
                            self:ScheduleTimer("SendChatMessage", self.db.global.scheduleTimer.sendMessageDelay*i, message, "SAY")
                        end
                        i = i + 1
                    end
                    del(messages)
                elseif type(messages) == "string" then
                    self:SendChatMessage(messages, "SAY")
                end
            end
        end

        del(group)
        del(near)
    end

    if self:TimeLeft(self.gratzTimerGuild) <= 0 and self:TimeLeft(self.gratzTimerNearGroup) <= 0 then
        wipe(self.gratzPendingList)
    end
    --[===[@debug@
    self:Printf("GratzForAchievement(%d): %f", flag or 0, (_G.GetTime() - start) )
    --@end-debug@]===]
end

function ArekosGratz:ACHIEVEMENT_EARNED(event, achievementID)
    if self:IAmAFK() and self.db.profile.protectAFK then return end
    local isGuildAchievement, wasEarnedByMe, earnedBy = select(12, _G.GetAchievementInfo(achievementID) )
    --[===[@debug@
    self:Printf("wasEarnedByMe: %s, isGuildAchievement: %s, earnedBy: %s", tostring(wasEarnedByMe), tostring(isGuildAchievement), tostring(earnedBy))
    --@end-debug@]===]
    if wasEarnedByMe and not isGuildAchievement then
        self:ListeningForCongratulations()
    end
end

function ArekosGratz:ListeningForCongratulations()
    if self:TimeLeft(self.thanksTimer) <= 0 then
        self.thanksTimer = self:ScheduleTimer("ThanksForCongratulations", random(self.db.profile.thanks.minWaitTime, self.db.profile.thanks.maxWaitTime), 0 )
    end
    if self.db.profile.thanks.guild then
        self:RegisterEvent("CHAT_MSG_GUILD", "CheckForCongratulations", MESSAGE_FLAGS.GUILD)
    end
    if self.db.profile.thanks.party then
        self:RegisterEvent("CHAT_MSG_PARTY", "CheckForCongratulations", MESSAGE_FLAGS.GROUP)
        self:RegisterEvent("CHAT_MSG_PARTY_LEADER", "CheckForCongratulations", MESSAGE_FLAGS.GROUP)
        self:RegisterEvent("CHAT_MSG_RAID", "CheckForCongratulations", MESSAGE_FLAGS.GROUP)
        self:RegisterEvent("CHAT_MSG_RAID_LEADER", "CheckForCongratulations", MESSAGE_FLAGS.GROUP)
    end
    if self.db.profile.thanks.instance then
        self:RegisterEvent("CHAT_MSG_INSTANCE_CHAT", "CheckForCongratulations", MESSAGE_FLAGS.INSTANCE)
        self:RegisterEvent("CHAT_MSG_INSTANCE_CHAT_LEADER", "CheckForCongratulations", MESSAGE_FLAGS.INSTANCE)
    end
    if self.db.profile.thanks.near ~= NEAR_TYPES.NONE then
        self:RegisterEvent("CHAT_MSG_SAY", "CheckForCongratulations", MESSAGE_FLAGS.NEAR)
        self:RegisterEvent("CHAT_MSG_WHISPER", "CheckForCongratulations", MESSAGE_FLAGS.NEAR)
    end
end

function ArekosGratz:UnregisterCongratulationsEvents()
    self:UnregisterEvent("CHAT_MSG_GUILD")
    self:UnregisterEvent("CHAT_MSG_PARTY")
    self:UnregisterEvent("CHAT_MSG_PARTY_LEADER")
    self:UnregisterEvent("CHAT_MSG_RAID")
    self:UnregisterEvent("CHAT_MSG_RAID_LEADER")
    self:UnregisterEvent("CHAT_MSG_SAY")
    self:UnregisterEvent("CHAT_MSG_WHISPER")
    self:UnregisterEvent("CHAT_MSG_INSTANCE_CHAT")
    self:UnregisterEvent("CHAT_MSG_INSTANCE_CHAT_LEADER")
end

do-- Congratulations checks
    local resultCache = setmetatable({}, {
        __mode = "k",
        __index = function(t,k)
            local found = false
            for trigger, inUse in pairs(ArekosGratz.db.global.gzTriggers) do
                if inUse then
                    if (" "..k.." "):find("%s"..trigger.."%s") then
                        found = true
                        break
                    end
                end
            end
            --[===[@debug@
            ArekosGratz:Printf("Result of %q: %s", k, tostring(found) )
            --@end-debug@]===]
            rawset(t,k,found)
            return found
        end,
    })

    function ArekosGratz:CheckForCongratulations(flag, event, message, playername)
        if type(playername) ~= "string" or playername:len() < 1 then return end
        local guildname = playername
        playername = lower( self:UniformPlayerName( playername ) )
        if self:IsMe(playername) or self:IsMyGuild(guildname) then return end
        --[===[@debug@
        self:Printf("Check player: %s", playername)
        --@end-debug@]===]
        if resultCache[ lower(message) ] then
            --[===[@debug@
            self:Printf("Add flag: %s", flag)
            --@end-debug@]===]
            self.thanksPendingList[playername] = bit.bor( (self.thanksPendingList[playername] or MESSAGE_FLAGS.NONE), flag )
        end
    end

    --@alpha@
    function ArekosGratz:TestResultCache(text)
        return resultCache[ lower(text) ]
    end
    --@end-alpha@
end

function ArekosGratz:ThanksForCongratulations(count)
    --[===[@debug@
    local start = _G.GetTime()
    --@end-debug@]===]
    local guild = get()
    local guildFound = 0
    local group = get()
    local groupFound = 0
    local instance = get()
    local instanceFound = 0
    local near  = get()
    local nearFound = 0

    local guildAllowed
    local groupAllowed
    local instanceAllowed
    local nearAllowed


    for name, flags in pairs(self.thanksPendingList) do
        if bit.band(flags, MESSAGE_FLAGS.GUILD) ~= 0 then
            if guildAllowed == nil then
                guildAllowed = self:HasChance(self.db.profile.chance.thanks.guild, MESSAGE_FLAGS.GUILD+MESSAGE_FLAGS.THANKS)
            end
            if guildAllowed then
                guild[name] = true
                guildFound = guildFound + 1
            end
        end
        if bit.band(flags, MESSAGE_FLAGS.GROUP) ~= 0 then
            if groupAllowed == nil then
                groupAllowed = self:HasChance(self.db.profile.chance.thanks.party, MESSAGE_FLAGS.GROUP+MESSAGE_FLAGS.THANKS)
            end
            if groupAllowed then
                group[name] = true
                groupFound = groupFound + 1
            end
        end
        if bit.band(flags, MESSAGE_FLAGS.INSTANCE) ~= 0 then
            if instanceAllowed == nil then
                instanceAllowed = self:HasChance(self.db.profile.chance.thanks.instance, MESSAGE_FLAGS.INSTANCE+MESSAGE_FLAGS.THANKS)
            end
            if instanceAllowed then
                instance[name] = true
                instanceFound = instanceFound + 1
            end
        end
        if bit.band(flags, MESSAGE_FLAGS.NEAR) ~= 0 then
            if nearAllowed == nil then
                nearAllowed = self:HasChance(self.db.profile.chance.thanks.near, MESSAGE_FLAGS.NEAR+MESSAGE_FLAGS.THANKS)
            end
            if nearAllowed then
                near[name] = true
                nearFound = nearFound + 1
            end
        end
    end
    if ( guildFound + groupFound + nearFound + instanceFound ) == 0 and (count or 0) < self.db.profile.thanks.maxExtraThanksTrys then
        --[===[@debug@
        self:Print("Retry: ", count or 0 )
        --@end-debug@]===]
        if self:TimeLeft(self.thanksTimer) <= 0 then
            --[===[@debug@
            self:Print("Start timer")
            --@end-debug@]===]
            self.thanksTimer = self:ScheduleTimer("ThanksForCongratulations", random(self.db.profile.thanks.minWaitTime, self.db.profile.thanks.maxWaitTime), ( (count or 0) + 1 ) )
        end

        -- Remove this tables or we have garbage lying around
        del(guild)
        del(group)
        del(instance)
        del(near)

        return
    else
        self:UnregisterCongratulationsEvents()
        wipe(self.thanksPendingList)
    end

    if guildFound > 0 then
        local messages = self:GetMessages(guild, "messageTHX", guildFound)
        if type(messages) == "table" then
            local i = 0
            for _, message in pairs(messages) do
                if i == 0 then
                    self:SendChatMessage(message, "GUILD")
                else
                    self:ScheduleTimer("SendChatMessage", self.db.global.scheduleTimer.sendMessageDelay*i, message, "GUILD" )
                end
                i = i + 1
            end
            del(messages)
        elseif type(messages) == "string" then
            self:SendChatMessage(messages, "GUILD")
        end
    end
    if groupFound > 0 then
        local channel = self:GetGroupChannel(true)

        if type(channel) == "string" and channel:len() > 1 then
            local messages = self:GetMessages(group, "messageTHX", groupFound)

            if type(messages) == "table" then
                local i = 0
                for _, message in pairs(messages) do
                    if i == 0 then
                        self:SendChatMessage(message, channel)
                    else
                        self:ScheduleTimer("SendChatMessage", self.db.global.scheduleTimer.sendMessageDelay*i, message, channel)
                    end
                    i = i + 1
                end
                del(messages)
            elseif type(messages) == "string" then
                self:SendChatMessage(messages, channel)
            end
        end
    end
    if instanceFound > 0 then
        local channel = "INSTANCE_CHAT"
        local messages = self:GetMessages(instance, "messageTHX", instanceFound)
        if type(messages) == "table" then
            local i = 0
            for _, message in pairs(messages) do
                if i == 0 then
                    self:SendChatMessage(message, channel)
                else
                    self:ScheduleTimer("SendChatMessage", self.db.global.scheduleTimer.sendMessageDelay*i, message, channel)
                end
                i = i + 1
            end
            del(messages)
        elseif type(messages) == "string" then
            self:SendChatMessage(messages, channel)
        end
    end
    if nearFound > 0 then
        if self.db.profile.thanks.near == NEAR_TYPES.WHISPER then
            for name in pairs(near) do
                local nameTable = get()
                nameTable[name] = true
                local messages = self:GetMessages(nameTable, "messageTHX", nearFound)
                if type(messages) == "table" then
                    local message = messages[random(1, #messages)]

                    self:SendChatMessage(message, "WHISPER", name)
                    del(messages)
                elseif type(messages) == "string" then
                    self:SendChatMessage(messages, "WHISPER", name)
                end
                del(nameTable)
            end
        elseif self.db.profile.thanks.near == NEAR_TYPES.SAY then
            local messages = self:GetMessages(near, "messageTHX", nearFound)
            if type(messages) == "table" then
                local i = 0
                for _, message in pairs(messages) do
                    if i == 0 then
                        self:SendChatMessage(message, "SAY")
                    else
                        self:ScheduleTimer("SendChatMessage", self.db.global.scheduleTimer.sendMessageDelay*i, message, "SAY")
                    end
                    i = i + 1
                end
                del(messages)
            elseif type(messages) == "string" then
                self:SendChatMessage(messages, "SAY")
            end
        end
    end

    -- clean up
    del(guild)
    del(group)
    del(instance)
    del(near)
    --[===[@debug@
    self:Printf("ThanksForCongratulations: %f", (_G.GetTime() - start) )
    --@end-debug@]===]
end

do-- GetMessage

    local lastMessages = {}
    local maxSize = 1
    local function cleanLastMessages(size)
        maxSize = size > maxSize and size or maxSize

        if #lastMessages > maxSize then
            for i=#lastMessages, maxSize, -1 do
                --@non-debug@
                tremove(lastMessages, maxSize)
                --@end-non-debug@
                --[===[@debug@
                ArekosGratz:Printf("Clean up: reaming=%d, message=%s", #lastMessages, tostring( tremove(lastMessages, maxSize) or "" ) )
                --@end-debug@]===]
            end
        end
    end

    local function getRandomMessage(list, notFirst)

        if ArekosGratz.db.profile.intellectual_random and not notFirst then
            ArekosGratz:ScheduleTimer(cleanLastMessages, ArekosGratz.db.global.scheduleTimer.cleanMessageDelay, #list)
        end

        if ArekosGratz.db.profile.intellectual_random and #list > 1 then
            local i = random(1, #list)
            local message = list[i]

            local found = false

            local preCount = #list
            preCount = preCount - ceil( preCount * 0.5 )

            for j=1, preCount do
                if lastMessages[j] == message then
                    found = true
                    break
                end
            end

            if found then
                tremove(list, i)
                return getRandomMessage(list, true)
            else
                tinsert(lastMessages, 1, message)
                return message
            end
        else
            local message = list[random(1, #list)]
            if ArekosGratz.db.profile.intellectual_random then
                tinsert(lastMessages, 1, message)
            end
            return message
        end
    end
    --@alpha@
    ArekosGratz.GetRandomMessage = getRandomMessage
    --@end-alpha@

    local function getUnusedNames(names, unusedNames)
        local count = 0
        unusedNames = unusedNames or get()

        for name, notUsed in pairs(names) do
            if notUsed then
                count = count + 1
                unusedNames[name] = true
            end
        end

        return unusedNames, count
    end

    -- Regexp to match the first character for UTF8 strings
    local MULTIBYTE_FIRST_CHAR = "^([\192-\255]?%a?[\128-\191]*)"

    local function modifyNames(list, removeRealm, capitalize)
        if #list > 0 then
            if removeRealm and capitalize then
                for i, name in ipairs(list) do
                    name = ("-"):split(name)

                    list[i] = name:gsub(MULTIBYTE_FIRST_CHAR, string.upper, 1)
                end
            elseif removeRealm then
                for i, name in ipairs(list) do
                    list[i] = ("-"):split(name)
                end
            elseif capitalize then
                for i, name in ipairs(list) do
                    local name, realm = ("-"):split(name)

                    if realm then
                        realm = realm:gsub(MULTIBYTE_FIRST_CHAR, string.upper, 1)
                        realm = "-"..realm
                    else
                        realm = ""
                    end

                    list[i] = name:gsub(MULTIBYTE_FIRST_CHAR, string.upper, 1)..realm
                end
            end
        else
            if removeRealm and capitalize then
                for name in pairs(list) do
                    list[name] = false -- mark as unused

                    name = ("-"):split(name)

                    name = name:gsub(MULTIBYTE_FIRST_CHAR, string.upper, 1)
                    list[name] = true
                end
            elseif removeRealm then
                for name in pairs(list) do
                    list[name] = false -- mark as unused

                    name = ("-"):split(name)
                    list[name] = true
                end
            elseif capitalize then
                for name in pairs(list) do
                    list[name] = false -- mark as unused

                    local name, realm = ("-"):split(name)

                    if realm then
                        realm = realm:gsub(MULTIBYTE_FIRST_CHAR, string.upper, 1)
                        realm = "-"..realm
                    else
                        realm = ""
                    end

                    name = name:gsub(MULTIBYTE_FIRST_CHAR, string.upper, 1)..realm
                    list[name] = true
                end
            end
        end

        return list
    end

    local function replaceNames(message, names, scope)
        if message:find("#n") and type(names) == "table" then

            names = modifyNames(names, ArekosGratz.db.profile[scope][message].removeRealmName, ArekosGratz.db.profile[scope][message].capitalizeFirstChar)

            if #names > 0 then
                return message:gsub("#n", strjoin(", ", unpack(names) ))
            else

                local namelist
                for name, inUse in pairs(names) do
                    if inUse then
                        if not namelist then
                            namelist = name
                        else
                            namelist = namelist .. ", " .. name
                        end
                    end
                end
                return message:gsub("#n", namelist)
            end
        else
            return message
        end
    end

    function ArekosGratz:GetMessages(namesList, scope, nameCounts)
        -- Priority list: onlyForAll > notForAll > onlyForAllMatch > notForAllMatch > forAll > filtered: onlyFor, notFor
        local onlyForAll        = get() -- [#] = message
        local notForAll         = get() -- [#] = message
        local onlyForAllMatch   = get() -- [#] = message
        local notForAllMatch    = get() -- [#] = message
        local forAll            = get() -- [#] = message
        local filtered          = get() -- [message] = { [#] = name }

        local filteredCount = 0
        local filteredMessage

        for message, data in pairs(self.db.profile[scope]) do
            if data.exists then
                --[===[@debug@
                self:Printf("Message: %q", message)
                --@end-debug@]===]
                if #data.onlyFor == 0 and #data.notFor == 0 then
                    tinsert(forAll, message)
                    --[===[@debug@
                    self:Printf("     forAll")
                    --@end-debug@]===]
                else
                    if #data.onlyFor > 0 then
                        local blocked = false

                        for _, rule in pairs(data.onlyFor) do
                            if rule:find('#') then
                                rule = self:GetToonNameFromBattleNetTag(rule)
                            end

                            if namesList[rule] then
                                if not filtered[message] then
                                    filtered[message] = get()
                                end
                                tinsert(filtered[message], rule)
                                --[===[@debug@
                                self:Printf("     filtered(only): %q", rule)
                                --@end-debug@]===]
                            else
                                blocked = true
                            end
                        end

                        if not blocked and ( type(filtered[message]) == "table" and nameCounts == #filtered[message] ) then
                            tinsert(onlyForAll, message)
                            --[===[@debug@
                            self:Printf("     onlyForAll")
                            --@end-debug@]===]
                        elseif ( type(filtered[message]) == "table" and nameCounts == #filtered[message] ) then
                            tinsert(onlyForAllMatch, message)
                            --[===[@debug@
                            self:Printf("     onlyForAllMatch")
                            --@end-debug@]===]
                        end
                    elseif #data.notFor > 0 then
                        local blocked = get()

                        for _, rule in pairs(data.notFor) do
                            if rule:find('#') then
                                rule = self:GetToonNameFromBattleNetTag(rule)
                            end

                            if namesList[rule] then
                                blocked[rule] = true
                            end
                        end

                        for name in pairs(namesList) do
                            if not blocked[name] then
                                if not filtered[message] then
                                    filtered[message] = get()
                                end
                                tinsert(filtered[message], name)
                                --[===[@debug@
                                self:Printf("     filtered(not): %q", name)
                                --@end-debug@]===]
                            end
                        end

                        if not next(blocked) and ( type(filtered[message]) == "table" and nameCounts == #filtered[message] ) then
                            tinsert(notForAll, message)
                            --[===[@debug@
                            self:Printf("     notForAll")
                            --@end-debug@]===]
                        elseif ( type(filtered[message]) == "table" and nameCounts == #filtered[message] ) then
                            --[===[@debug@
                            self:Printf("     notForAllMatch")
                            --@end-debug@]===]
                        end
                        del(blocked)
                    end

                    if type(filtered[message]) == "table" and #filtered[message] > filteredCount then
                        filteredMessage = message
                        filteredCount = #filtered[message]
                        --[===[@debug@
                        self:Printf("     filtered-max(%d): %q", filteredCount, message)
                        --@end-debug@]===]
                    end
                end
            end
        end

        local returnMessage

        --[===[@debug@
        self:Printf("Result: ")
        --@end-debug@]===]

        if #onlyForAll > 0 then
            --[===[@debug@
            self:Printf("     onlyForAll")
            --@end-debug@]===]
            local message = getRandomMessage(onlyForAll)

            returnMessage = replaceNames(message, namesList, scope)
        elseif #notForAll > 0 then
            --[===[@debug@
            self:Printf("     notForAll")
            --@end-debug@]===]
            local message = getRandomMessage(notForAll)

            returnMessage = replaceNames(message, namesList, scope)
        elseif #onlyForAllMatch > 0 then
            --[===[@debug@
            self:Printf("     onlyForAllMatch")
            --@end-debug@]===]
            local message = getRandomMessage(onlyForAllMatch)

            returnMessage = replaceNames(message, namesList, scope)
        elseif #notForAllMatch > 0 then
            --[===[@debug@
            self:Printf("     notForAllMatch")
            --@end-debug@]===]
            local message = getRandomMessage(notForAllMatch)

            returnMessage = replaceNames(message, namesList, scope)
        elseif #forAll > 0 then
            --[===[@debug@
            self:Printf("     forAll")
            --@end-debug@]===]
            local message = getRandomMessage(forAll)

            returnMessage = replaceNames(message, namesList, scope)
        elseif next(filtered) and filteredMessage then
            -- Create a list with only used names.
            -- This why we don't run in the problem to send messages for player with no matching messages
            local namesList = {}
            for message, names in pairs(filtered) do
                for i, name in ipairs(names) do
                    namesList[name] = true
                end
            end

            --[===[@debug@
            self:Printf("     filtered")
            --@end-debug@]===]
            returnMessage = get()

            tinsert(returnMessage, replaceNames(filteredMessage, filtered[filteredMessage], scope) )

            for i, name in ipairs(filtered[filteredMessage]) do
                namesList[name] = false
            end
            del(filtered[filteredMessage])
            filtered[filteredMessage] = nil

            local unusedNames, unusedNamesCount = getUnusedNames(namesList)
            del(namesList)

            while true do -- breaks out when we have no message remaining or none unused names
                local matchCount = 0
                local matchMessage
                for message, data in pairs(filtered) do
                    local count = 0
                    for i, name in ipairs(data) do
                        if unusedNames[name] then
                            count = count + 1
                        end
                    end
                    if count > matchCount then
                        matchCount = count
                        matchMessage = message

                        if count >= unusedNamesCount then
                            break
                        end
                    end
                end

                if matchMessage then
                    local usedNames = get()

                    for i, name in ipairs(filtered[matchMessage]) do
                        if unusedNames[name] then
                            unusedNames[name] = false
                            tinsert(usedNames, name)
                        end
                    end

                    tinsert(returnMessage, replaceNames(matchMessage, usedNames, scope) )
                    del(usedNames)

                    if matchCount >= unusedNamesCount then
                        break
                    end

                    del(filtered[matchMessage])
                    filtered[matchMessage] = nil
                    if not next(filtered) then
                        break
                    end

                    local _unusedNames
                    _unusedNames, unusedNamesCount = getUnusedNames(unusedNames)
                    del(unusedNames)
                    unusedNames = _unusedNames
                else
                    break
                end
            end

            del(unusedNames)
        end

        -- clean up
        for k in pairs(filtered) do
            del(filtered[k])
            filtered[k] = nil
        end
        del(filtered)
        del(onlyForAll)
        del(notForAll)
        del(onlyForAllMatch)
        del(notForAllMatch)
        del(forAll)

        return returnMessage or get()
    end
end

function ArekosGratz:UniformPlayerName(playername, realmOverride)
    playername = playername:gsub("%s+", "")
    local name, realm = ("-"):split( playername )

    if type(realmOverride) == "string" and realmOverride:len() > 0 then
        realm = realmOverride:gsub("%s+", "")
    elseif type(realm) ~= "string" or realm:len() < 1 then
        realm = HOMEREALM
    end

    return name.."-"..realm, name, realm
end
function ArekosGratz:UniformGuildName(guildname, realmOverride)
    local guild, realm = ("-"):split(guildname)

    if type(realmOverride) == "string" and realmOverride:trim():len() > 0 then
        realm = realmOverride:gsub("%s+", "")
    else
        if type(realm) ~= "string" then
                realm = HOMEREALM
        else
            realm = realm:gsub("%s+", "")
            if realm:len() < 1 then
                realm = HOMEREALM
            end
        end
    end

    return guild.."-"..realm, guild, realm
end

function ArekosGratz:SendChatMessage(message, channel, to)
    if self:IAmAFK() and self.db.profile.protectAFK then return end
    if type(message) ~= "string" or message:len() < 1 or type(channel) ~= "string" or channel:len() < 1 then return end

    --[===[@debug@
    self:Printf("SCM: [%s] %s", to or channel, message)
    --@end-debug@]===]

    if self.db.profile.playSound then
        local soundFile = LSM:Fetch(LSM.MediaType.SOUND, self.db.profile.sound)
        _G.PlaySoundFile(soundFile, self.db.profile.soundChannel)
    end

    if to then
        _G.SendChatMessage(message, channel, nil, to)
    else
        _G.SendChatMessage(message, channel)
    end
end


local optionsHandler = setmetatable({ cache = {} },{
    __index = function(t, k)
        if k == "db" then
            return ArekosGratz.db
        else
            return rawget(t,k)
        end
    end
})

function ArekosGratz:ResetOptionsCache()
    wipe(optionsHandler.cache)
end

local function getInternCacheKey(info)
    local key = ""
    if type(info) == "table" then
        for i=1, #info do
            key = key..( info[i] or "root"..i )
        end
    else
        key = tostring(info)
    end
    return key
end
function optionsHandler:SetInternCache(info, value)
    local key = getInternCacheKey(info)
    self.cache[key] = value
end
function optionsHandler:GetInternCache(info, defaultValue)
    local defaultValue = defaultValue == nil and "" or defaultValue
    local key = getInternCacheKey(info)
    local value = self.cache[key]

    if value == nil then
        value = defaultValue
    end
    return value
end
function optionsHandler:SetCache(info, value)
    self:SetInternCache(info, value)
end
function optionsHandler:GetCache(info)
    return self:GetInternCache(info, "")
end

function optionsHandler:Set(info, value)
    local scope = info.arg or "profile"
    local name = info[#info]

    self.db[scope][name] = value
end
function optionsHandler:Get(info)
    local scope = info.arg or "profile"
    local name = info[#info]

    return self.db[scope][name]
end
function optionsHandler:IsIgnoreInstanceGroupDisabled(info)
    return not self.db.profile.congratulations.party
end
function optionsHandler:SetCongratulations(info, value)
    local name = info[#info]

    self.db.profile.congratulations[name] = value
end
function optionsHandler:GetCongratulations(info)
    local name = info[#info]

    return self.db.profile.congratulations[name]
end
function optionsHandler:SetThanks(info, value)
    local name = info[#info]

    self.db.profile.thanks[name] = value
end
function optionsHandler:GetThanks(info)
    local name = info[#info]

    return self.db.profile.thanks[name]
end
function optionsHandler:SetMinWaitTime(info, value)
    local scope
    if type(info.arg) == "string" and info.arg:len() > 0 then
        scope = info.arg
    else
        return
    end

    if self.db.profile[scope].maxWaitTime > value then
        self.db.profile[scope].minWaitTime = value
    end
end
function optionsHandler:SetMaxWaitTime(info, value)
    local scope
    if type(info.arg) == "string" and info.arg:len() > 0 then
        scope = info.arg
    else
        return
    end

    if self.db.profile[scope].minWaitTime < value then
        self.db.profile[scope].maxWaitTime = value
    end
end

local ID_KEY = "MK:"
local ID_KEY_LENGTH = ID_KEY:len()

function optionsHandler:ValidateAddMessage(info, value, scope)
    local scope = type(scope) == "string" and scope or info[#info-1] or info.arg
    local name = info[#info]

    if name:sub(-6) == "Button" then
        info[#info] = name:sub(0,-7)
        value = self:GetInternCache(info)
        info[#info] = name
    end

    local result = true

    if value:trim():len() < 1 then
        result = L["Message is empty."]
    elseif self.db.profile[scope][value].exists then
        result = L["Exists already."]
    end

    if info.uiType == "dialog" and type(result) == "string" then
        ArekosGratz:Printf("%s: %s", info.option.name or L["Input"], result)
    end
    return result
end

function optionsHandler:AddMessage(info, value)
    local scope = info[#info-1] or info.arg
    local name = info[#info]

    local validated = self:ValidateAddMessage(info, value)
    if type(validated) == "string" or not validated then
        return
    end

    if name:sub(-6) == "Button" then
        info[#info] = name:sub(0,-7)
        value = self:GetInternCache(info)
        self:SetInternCache(info, "")
        info[#info] = name
    end

    self.db.profile[scope][value:trim()].exists = true
end
function optionsHandler:DeleteMessage(info, value)
    local scope = info[#info-2] or info.arg
    local key = info[#info-1]:sub(ID_KEY_LENGTH+1)

    self.db.profile[scope][key] = nil

    info.option = nil
end
function optionsHandler:HasNoneNameReplace(info)
    local key = info[#info-1]:sub(ID_KEY_LENGTH+1)

    local cache = self:GetInternCache(info, 1)
    if cache == 1 then
        if key:find("#n") then
            cache = false
        else
            cache = true
        end
        self:SetInternCache(info, cache)
    end
    return cache
end
function optionsHandler:SetMessage(info, value)
    local scope = info.arg or info[#info-2]
    local key = info[#info-1]:sub(ID_KEY_LENGTH+1)
    local name = info[#info]

    self.db.profile[scope][key][name] = value
end
function optionsHandler:GetMessage(info)
    local scope = info.arg or info[#info-2]
    local key = info[#info-1]:sub(ID_KEY_LENGTH+1)
    local name = info[#info]

    return self.db.profile[scope][key][name]
end
function optionsHandler:GetMessageKey(info)
    return self:GetInternCache(info, false) or info[#info-1]:sub(ID_KEY_LENGTH+1)
end
function optionsHandler:RenameMessage(info, value)
    local scope = info[#info-2] or info.arg
    local key = info[#info-1]:sub(ID_KEY_LENGTH+1)
    local name = info[#info]

    local validated = self:ValidateAddMessage(info, value, scope)
    if type(validated) == "string" or not validated then
        return
    end

    if name:sub(-6) == "Button" then
        info[#info] = name:sub(0,-7)
        value = self:GetInternCache(info, key)
        info[#info] = name
    end

    value = value:trim()

    self.db.profile[scope][value] = self.db.profile[scope][key]
    if value ~= key then
        self.db.profile[scope][key] = nil
    end
end

function optionsHandler:SetMessageFilter(info, value)
    local scope = info[#info-2] or info.arg
    local key = info[#info-1]:sub(ID_KEY_LENGTH+1)
    local name = info[#info]

    local hasBattleNetTag = false
    local values = get()
    for v in value:gmatch("[^\n]+") do
        v = v:gsub("%s+", "")

        if v:len() > 1 then
            if v:find('#') then
                local name, id = ('#'):split(v)

                if type(id) == "string" and id:len() == 4 and name:len() > 1 then
                    tinsert(values, lower(name..'#'..id) )

                    hasBattleNetTag = true
                end
            else
                local name = ArekosGratz:UniformPlayerName(v)
                tinsert(values, lower(name) )
            end
        end
    end

    del(self.db.profile[scope][key][name])
    self.db.profile[scope][key][name] = values
    self.db.profile[scope][key][name.."_HasBattleNetTag"] = hasBattleNetTag

    if hasBattleNetTag and not ArekosGratz:HasBattleNetTags() then
        ArekosGratz:CheckForBattleNetTags()

        if ArekosGratz:HasBattleNetTags() then
            ArekosGratz:InitializeBattleNetTagIndex()
        end
    elseif not hasBattleNetTag and ArekosGratz:HasBattleNetTags() then
        ArekosGratz:CheckForBattleNetTags()

        if not ArekosGratz:HasBattleNetTags() then
            ArekosGratz:DisableBattleNetTagIndex()
        end
    end

    self:SetInternCache(info, false)
end
function optionsHandler:GetMessageFilter(info)
    local cache = self:GetInternCache(info, false)
    if not cache then
        local scope = info.arg or info[#info-2]
        local key = info[#info-1]:sub(ID_KEY_LENGTH+1)
        local name = info[#info]

        local data = self.db.profile[scope][key][name]
        cache = strjoin("\n", unpack(data))
        self:SetInternCache(info, cache)
    end

    return cache
end

function optionsHandler:SetTrigger(info, value)
    local triggers = self.db.global.gzTriggers

    for trigger, inUse in pairs(triggers) do
        if inUse then
            triggers[trigger] = false
        end
    end
    for trigger in value:gmatch("[^\n]+") do
        triggers[lower( trigger:trim() )] = true
    end
    self:SetInternCache(info, false)
end
function optionsHandler:GetTrigger(info)
    local cache = self:GetInternCache(info, false)
    if not cache then
        local data = get()

        for trigger, inUse in pairs(self.db.global.gzTriggers) do
            if inUse then
                tinsert(data, trigger)
            end
        end

        cache = strjoin("\n", unpack(data) )
        del(data)
        self:SetInternCache(info, cache)
    end

    return cache
end

function optionsHandler:SetChance(info, value)
    local name = info[#info]
    local scope = info.arg

    self.db.profile.chance[scope][name] = value
end
function optionsHandler:GetChance(info)
    local name = info[#info]
    local scope = info.arg

    return self.db.profile.chance[scope][name]
end
function optionsHandler:IsChanceDisabled(info)
    local name = info[#info]
    local scope = info.arg

    local value = self.db.profile[scope][name]

    if type(value) == "boolean" then
        return not value
    else
       return value == NEAR_TYPES.NONE
    end
end


local configs = {
    type    = "group",
    name    = L["Areko's Gratz"],
    handler = optionsHandler,
    set     = "Set",
    get     = "Get",
    args = {
        main = {
            type    = "group",
            name    = L["Areko's Gratz"],
            handler = optionsHandler,
            set     = "Set",
            get     = "Get",
            args = {
                protectAFK = {
                    type    = "toggle",
                    name    = L["Protect AFK-Status"],
                    desc    = L["No congratulation while AFK"],
                    order   = 1,
                },
                playSound = {
                    type    = "toggle",
                    name    = L["Play Sound"],
                    desc    = L["Play a Sound when sending a message."],
                    order   = 1.1,
                },
                sound = {
                    type            = "select",
                    dialogControl   = "LSM30_Sound",
                    name            = L["Sound"],
                    desc            = L["Sound which should be played."],
                    values          = LSM:HashTable(LSM.MediaType.SOUND),
                    order           = 1.1,
                },
                congratulations = {
                    type    = "group",
                    name    = L["Congratulate"],
                    desc    = L["Settings for congratulations"],
                    order   = 2,
                    set     = "SetCongratulations",
                    get     = "GetCongratulations",
                    inline  = true,
                    args    = {
                        guild = {
                            type    = "toggle",
                            name    = L["Guild"],
                            desc    = L["Congratulations to the guild channel"],
                            order   = 1,
                            arg     = "congratulations",
                        },
                        party = {
                            type    = "toggle",
                            name    = L["Party"],
                            desc    = L["Congratulations to the party channel"],
                            order   = 2,
                        },
                        ignoreInstanceGroup = {
                            type        = "toggle",
                            name        = L["Ignore Instance"],
                            desc        = L["Don't congratulate to instance groups"],
                            disabled    = "IsIgnoreInstanceGroupDisabled",
                            order       = 2.5,
                        },
                        near = {
                            type    = "select",
                            name    = L["Congratulations to near players"],
                            values  = NEAR_TYPES_VALUES,
                            style   = "radio",
                            order   = 3
                        },
                        minWaitTime = {
                            type    = "range",
                            name    = L["Min. wait time to congratulation"],
                            min     = 0,
                            max     = 60,
                            step    = 1,
                            bigStep = 1,
                            arg     = "congratulations",
                            set     = "SetMinWaitTime",
                            order   = 4,
                            width   = "double",
                        },
                        maxWaitTime = {
                            type    = "range",
                            name    = L["Max. wait time to congratulation"],
                            min     = 0,
                            max     = 60,
                            step    = 1,
                            bigStep = 1,
                            arg     = "congratulations",
                            set     = "SetMaxWaitTime",
                            order   = 5,
                            width   = "double",
                        },
                        guildInOtherChannel = {
                            type    = "toggle",
                            name    = L["Congratulate guildmembers over other channels"],
                            desc    = L["Congratulations to guildmembers who in our party or near your in other channels as guildchannel"],
                            order   = 7,
                            width   = "full"
                        },
                        chance = {
                            type    = "group",
                            name    = L["Congratulations chance"],
                            set     = "SetChance",
                            get     = "GetChance",
                            disabled= "IsChanceDisabled",
                            order   = 6,
                            args    = {
                                guild = {
                                    type    = "range",
                                    name    = L["Guild"],
                                    desc    = L["Chance to congratulate in guild."],
                                    min     = 0,
                                    max     = 1,
                                    bigStep = 0.05,
                                    step    = 0.01,
                                    isPercent = true,
                                    arg     = "congratulations",
                                    order   = 1,
                                },
                                party = {
                                    type    = "range",
                                    name    = L["Party"],
                                    desc    = L["Chance to congratulate in party channels."],
                                    min     = 0,
                                    max     = 1,
                                    bigStep = 0.05,
                                    step    = 0.01,
                                    isPercent = true,
                                    arg     = "congratulations",
                                    order   = 2,
                                },
                                near = {
                                    type    = "range",
                                    name    = L["Near"],
                                    desc    = L["Chance to congratulate player near you."],
                                    min     = 0,
                                    max     = 1,
                                    bigStep = 0.05,
                                    step    = 0.01,
                                    isPercent = true,
                                    arg     = "congratulations",
                                    order   = 3,
                                },
                            },
                        },
                    },
                },
                thanks = {
                    type    = "group",
                    name    = L["Thanks"],
                    desc    = L["Settings for saying thanks"],
                    set     = "SetThanks",
                    get     = "GetThanks",
                    inline  = true,
                    order   = 3,
                    args    = {
                        guild = {
                            type    = "toggle",
                            name    = L["Guild"],
                            desc    = L["Saying thanks to the guild channel"],
                            order   = 8,
                        },
                        party = {
                            type    = "toggle",
                            name    = L["Party"],
                            desc    = L["Saying thanks to the party channel"],
                            order   = 8,
                        },
                        instance = {
                            type    = "toggle",
                            name    = L["Instance"],
                            desc    = L["Saying thanks to the instance group"],
                            order   = 8.5,
                        },
                        near = {
                            type    = "select",
                            name    = L["Saying thanks near players"],
                            values  = NEAR_TYPES_VALUES,
                            style   = "radio",
                            order   = 9,
                        },
                        minWaitTime = {
                            type    = "range",
                            name    = L["Min. wait time to thanks"],
                            min     = 0,
                            max     = 60,
                            step    = 1,
                            bigStep = 1,
                            arg     = "thanks",
                            set     = "SetMinWaitTime",
                            order   = 10,
                            width   = "double",
                        },
                        maxWaitTime = {
                            type    = "range",
                            name    = L["Max. wait time to thanks"],
                            min     = 0,
                            max     = 60,
                            step    = 1,
                            bigStep = 1,
                            arg     = "thanks",
                            set     = "SetMaxWaitTime",
                            order   = 11,
                            width   = "double",
                        },
                        maxExtraThanksTrys = {
                            type    = "range",
                            name    = L["Max. times to retry to thanks for congratulations"],
                            min     = 0,
                            softMax = 3,
                            max     = 10,
                            bigStep = 1,
                            arg     = "thanks",
                            order   = 12,
                        },
                        chance = {
                            type    = "group",
                            name    = L["Thanks chance"],
                            set     = "SetChance",
                            get     = "GetChance",
                            disabled= "IsChanceDisabled",
                            order   = 13,
                            args    = {
                                guild = {
                                    type    = "range",
                                    name    = L["Guild"],
                                    desc    = L["Chance to thanks in guild."],
                                    min     = 0,
                                    max     = 1,
                                    bigStep = 0.05,
                                    step    = 0.01,
                                    isPercent = true,
                                    arg     = "thanks",
                                    order   = 1,
                                },
                                party = {
                                    type    = "range",
                                    name    = L["Party"],
                                    desc    = L["Chance to thanks in party channels."],
                                    min     = 0,
                                    max     = 1,
                                    bigStep = 0.05,
                                    step    = 0.01,
                                    isPercent = true,
                                    arg     = "thanks",
                                    order   = 2,
                                },
                                near = {
                                    type    = "range",
                                    name    = L["Near"],
                                    desc    = L["Chance to thanks player near you."],
                                    min     = 0,
                                    max     = 1,
                                    bigStep = 0.05,
                                    step    = 0.01,
                                    isPercent = true,
                                    arg     = "thanks",
                                    order   = 3,
                                },
                                instance = {
                                    type    = "range",
                                    name    = L["Instance"],
                                    desc    = L["Chance to thanks in instance groups."],
                                    min     = 0,
                                    max     = 1,
                                    bigStep = 0.05,
                                    step    = 0.01,
                                    isPercent = true,
                                    arg     = "thanks",
                                    order   = 4,
                                },
                            },
                        },
                    },
                },
            },
        },
        messageGZ = {
            type        = "group",
            name        = L["Congratulate"],
            childGroups = "tree",
            handler     = optionsHandler,
            set         = "Set",
            get         = "Get",
            args = {
                AddMessageGZ = {
                    name        = L["Congratulations message"],
                    desc        = L["Use #n for playername"],
                    type        = "input",
                    arg         = "messageGZ",
                    set         = "SetCache",
                    get         = "GetCache",
                    order       = 1,
                },
                AddMessageGZButton = {
                    name        = L["Add message"],
                    type        = "execute",
                    arg         = "messageGZ",
                    validate    = "ValidateAddMessage",
                    func        = "AddMessage",
                    order       = 1.1,
                },
            },
        },
        messageTHX = {
            type        = "group",
            name        = L["Thanks"],
            childGroups = "tree",
            handler     = optionsHandler,
            set         = "Set",
            get         = "Get",
            args = {
                AddMessageTHX = {
                    name        = L["Thanks message"],
                    desc        = L["Use #n for playername"],
                    type        = "input",
                    arg         = "messageTHX",
                    set         = "SetCache",
                    get         = "GetCache",
                    order       = 1,
                },
                AddMessageTHXButton = {
                    name        = L["Add message"],
                    type        = "execute",
                    arg         = "messageTHX",
                    validate    = "ValidateAddMessage",
                    func        = "AddMessage",
                    order       = 1.1,
                },
            },
        },
        trigger = {
            type = "group",
            name = L["Thanks Triggers"],
            arg  = "global",
            handler = optionsHandler,
            set  = "SetTrigger",
            get  = "GetTrigger",
            args = {
                list = {
                    name    = L["Congratulations Keywords"],
                    desc    = L["List of triggers for thanks\nOne word per line"],
                    type    = "input",
                    multiline = 20,
                    width   = "full",
                    order   = 1
                    },
                },
        },
    },
}

do
    local filterDescription = L["One player per line.\n\nName-Realm: Match on playername\nName#0000: Battle.net Tag"]
    function ArekosGratz:GetMessageOptionsTable(uiType, uiName, appName, scope)
        local message = configs.args[scope].args

        for key, data in pairs(self.db.profile[scope]) do
            if not data.isIntern then
                local configKey = ID_KEY..key

                message[configKey] = message[configKey] or {}
                message[configKey].name    = key
                message[configKey].type    = "group"
                message[configKey].set     = "SetMessage"
                message[configKey].get     = "GetMessage"
                message[configKey].arg     = scope
                message[configKey].args    = message[configKey].args or {}

                message[configKey].args.rename = message[configKey].args.rename or {}
                message[configKey].args.rename.name    = L["Rename message"]
                message[configKey].args.rename.desc    = L["Use #n for playername"]
                message[configKey].args.rename.type    = "input"
                message[configKey].args.rename.set     = "SetCache"
                message[configKey].args.rename.get     = "GetMessageKey"
                message[configKey].args.rename.arg     = scope
                message[configKey].args.rename.order   = 1

                message[configKey].args.renameButton = message[configKey].args.renameButton or {}
                message[configKey].args.renameButton.name      = L["Rename message"]
                message[configKey].args.renameButton.type      = "execute"
                message[configKey].args.renameButton.validate  = "ValidateAddMessage"
                message[configKey].args.renameButton.func      = "RenameMessage"
                message[configKey].args.renameButton.arg       = scope
                message[configKey].args.renameButton.order     = 2

                message[configKey].args.removeRealmName = message[configKey].args.removeRealmName or {}
                message[configKey].args.removeRealmName.name      = L["Remove Realm"]
                message[configKey].args.removeRealmName.desc      = L["Remove Realm from the playername"]
                message[configKey].args.removeRealmName.type      = "toggle"
                message[configKey].args.removeRealmName.arg       = scope
                message[configKey].args.removeRealmName.hidden    = "HasNoneNameReplace"
                message[configKey].args.removeRealmName.order     = 2.1

                message[configKey].args.capitalizeFirstChar = message[configKey].args.capitalizeFirstChar or {}
                message[configKey].args.capitalizeFirstChar.name    = L["Capitalize Name"]
                message[configKey].args.capitalizeFirstChar.desc    = L["Capitalize the first char from the playername"]
                message[configKey].args.capitalizeFirstChar.type    = "toggle"
                message[configKey].args.capitalizeFirstChar.arg     = scope
                message[configKey].args.capitalizeFirstChar.hidden  = "HasNoneNameReplace"
                message[configKey].args.capitalizeFirstChar.order   = 2.2

                message[configKey].args.onlyFor = message[configKey].args.onlyFor or {}
                message[configKey].args.onlyFor.name       = L["Only for:"]
                message[configKey].args.onlyFor.desc       = filterDescription
                message[configKey].args.onlyFor.type       = "input"
                message[configKey].args.onlyFor.multiline  = 9
                message[configKey].args.onlyFor.set        = "SetMessageFilter"
                message[configKey].args.onlyFor.get        = "GetMessageFilter"
                message[configKey].args.onlyFor.width      = "double"
                message[configKey].args.onlyFor.arg        = scope
                message[configKey].args.onlyFor.order      = 3

                message[configKey].args.notFor = message[configKey].args.notFor or {}
                message[configKey].args.notFor.name        = L["Not for:"]
                message[configKey].args.notFor.desc        = filterDescription
                message[configKey].args.notFor.type        = "input"
                message[configKey].args.notFor.multiline   = 9
                message[configKey].args.notFor.set         = "SetMessageFilter"
                message[configKey].args.notFor.get         = "GetMessageFilter"
                message[configKey].args.notFor.width       = "double"
                message[configKey].args.notFor.arg         = scope
                message[configKey].args.notFor.order       = 4

                message[configKey].args.deleteButton = message[configKey].args.deleteButton or {}
                message[configKey].args.deleteButton.name           = L["Delete"]
                message[configKey].args.deleteButton.type           = "execute"
                message[configKey].args.deleteButton.func           = "DeleteMessage"
                message[configKey].args.deleteButton.confirm        = true
                message[configKey].args.deleteButton.confirmText    = L["Delete %q message?"]:format(key)
                message[configKey].args.deleteButton.arg            = scope
                message[configKey].args.deleteButton.order          = 5
            end
        end

        for key in pairs(message) do
            if key:sub(0,ID_KEY_LENGTH) == ID_KEY then
                if self.db.profile[scope] and not self.db.profile[scope][key:sub(ID_KEY_LENGTH+1)].exists then
                    message[key] = nil
                end
            end
        end

        return configs.args[scope]
    end
    function ArekosGratz.GetMessageGZOptionsTable(uiType, uiName, appName)
        return ArekosGratz:GetMessageOptionsTable(uiType, uiName, appName, "messageGZ")
    end
    function ArekosGratz.GetMessageTHXOptionsTable(uiType, uiName, appName)
        return ArekosGratz:GetMessageOptionsTable(uiType, uiName, appName, "messageTHX")
    end
end

function ArekosGratz:RegisterConfig()
    local registry = LibStub("AceConfigRegistry-3.0")
    local dialog = LibStub("AceConfigDialog-3.0")
    local skip = true
    --[===[@debug@
    skip = false
    --@end-debug@]===]

    registry:RegisterOptionsTable(ADDONNAME.."OptionsMain", configs.args.main, skip)
    registry:RegisterOptionsTable(ADDONNAME.."OptionsMessageGZ", self.GetMessageGZOptionsTable, skip)
    registry:RegisterOptionsTable(ADDONNAME.."OptionsMessageTHX", self.GetMessageTHXOptionsTable, skip)
    registry:RegisterOptionsTable(ADDONNAME.."OptionsTrigger", configs.args.trigger, skip)

    local profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
    registry:RegisterOptionsTable(ADDONNAME.."Profile",  profiles)


    dialog:AddToBlizOptions(ADDONNAME.."OptionsMain", L["Areko's Gratz"], nil)
    dialog:AddToBlizOptions(ADDONNAME.."OptionsMessageGZ", L["Congratulate"], L["Areko's Gratz"])
    dialog:AddToBlizOptions(ADDONNAME.."OptionsMessageTHX", L["Thanks"], L["Areko's Gratz"])
    dialog:AddToBlizOptions(ADDONNAME.."OptionsTrigger", L["Thanks Triggers"], L["Areko's Gratz"])

    --L["Profile"]
    dialog:AddToBlizOptions(ADDONNAME.."Profile", profiles.name, L["Areko's Gratz"])
end


